#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vga.h>
#include "tetX.h"

extern long  Ticks;
extern int Level, JoyEnabled, ShowNext, Demo;
extern int ScreenWidth, ScanWidth, ScreenHeight;
extern int ScreenDepth, VideoChipset;

extern char FillPats[10][8];
extern char BgColourName[];
extern int BkgrPatColour[10];
extern int ScrollerBusy, ScrollAvailable;
extern int TimerSuspended;
extern IMAGE_HEADER Intelligent_Hdr, tetris_Hdr;

int ScreenBusy;
int BoardX, BoardY, BoardW, BoardH;
int JoyX, JoyY, JoyW, JoyH;
int ScoreX, ScoreY, ScoreW, ScoreH;
int NextX, NextY, NextW, NextH;
int PictureX, PictureY, PictureW, PictureH;
int FgrColour, BkgrColour, TxtColour, HiliteColour;

unsigned char *ScreenContext = NULL;
static int  HdrX, HdrY, HdrW, HdrH;


//-------------------------------------------------------------
void InitBoard(void)
{  int BrdRight;


   BoardW = BRD_WIDTH * FIG_WIDTH;
   BoardH = BRD_HEIGHT * FIG_HEIGHT;

   BoardY = (ScreenHeight - BoardH) >> 1;
   if (ScreenHeight <= 350) BoardY = 3;
                      else  BoardY += 8;
   BoardX = (ScreenWidth - BoardW) >> 1;

   // Calculate right window position
   BrdRight = BoardX + BoardW;

   ScoreW = TABLE_WIDTH*8 + 16;
   ScoreX = (ScreenWidth + BrdRight - ScoreW) >> 1;
   ScoreY = BoardY;                        // Top alignment
   ScoreH = 88;

   NextX = ScoreX;
   NextW = ScoreW;
   NextH = 92;
   NextY = BoardY + ((BoardH-NextH) >> 1);   // Central alignment


   JoyX = ScoreX;
   JoyW = ScoreW;
   JoyH = 88;
   JoyY = BoardY + BoardH - JoyH;         // Bottom alignment

   PictureW = 160;
   PictureX = (BoardX - PictureW) >> 1;
   PictureH = BoardH;
   PictureY = BoardY;

   BkgrColour = BKGR_COLOUR;
   TxtColour =  TEXT_COLOUR;
   HiliteColour = HILITE_COLOUR;

   setPalette(&tetris_Hdr, 16);

   if (ScreenHeight > 350)
   { HdrW = tetris_Hdr.imwid;
     HdrH = tetris_Hdr.imhei;
     HdrX = (ScreenWidth - HdrW)>>1;
     HdrY = (BoardY - HdrH) >> 1;
   }

    if (strlen(BgColourName) > 0 )
                    SetColour(BkgrColour, BgColourName);

   ScreenBusy = 0;
}

//-------------------------------------------------------------

void setrow(int x, int y, int wid, unsigned char *row)
{  if (ScreenContext)
   {  unsigned char *ptr = ScreenContext + (ScreenWidth * y + x);
      memcpy(ptr, row, wid);
   }
   else
   { 
     while ((ScreenBusy || ScrollerBusy) && !TimerSuspended) pause(); 
     ScreenBusy = 1;
     vga_drawscansegment(row, x, y, wid);
     ScreenBusy = 0;
   }
}

int  AllocScreenContext(void)
{
   if (ScreenContext) free (ScreenContext);
   ScreenContext = (unsigned char *)calloc(ScreenHeight, ScreenWidth);

   return ScreenContext ? 0: 4;
}

int  ScreenContextToScreen(void)
{

   if (!ScreenContext) return 4;


   while((ScreenBusy || ScrollerBusy) && !TimerSuspended) pause();
   ScreenBusy = 1;

   SuspendScroller();
   vga_waitretrace();

   if (ScreenDepth == 4)
       vga_copytoplanar16(ScreenContext, ScreenWidth,
               0, ScreenWidth>>3, ScreenWidth, ScreenHeight);
   else
   {  char *ptr = ScreenContext;
      int y;

      for(y=0; y<ScreenHeight; y++)
      {  vga_drawscansegment(ptr, 0, y, ScreenWidth);
        ptr += ScreenWidth;
      }
   }

   ResumeScroller();
   ScreenBusy = 0;
   return 0;
}

int  ContextRectToScreen(RECT *rc)
{  int x, y, w, toy;
   char *ptr = ScreenContext;

   if (!ptr) return 4;

   while((ScreenBusy || ScrollerBusy) && !TimerSuspended) pause();
   ScreenBusy = 1;

   SuspendScroller();

   x = rc->left;
   w = rc->right - x;
   toy = rc->bottom;
   y = rc->top;
   ptr = ptr + y * ScreenWidth + x;

   while(y<=toy)
   {
      vga_drawscansegment(ptr, x, y++, w);
      ptr += ScreenWidth;
   }

   ResumeScroller();
   ScreenBusy = 0;
   return 0;
}


int  ScreenToScreenContext(void)
{
   char *ptr = ScreenContext;
   int y;

   if (!ptr) return 4;

   SuspendScroller();
   for(y=0; y<ScreenHeight; y++)
   {  vga_getscansegment(ptr, 0, y, ScreenWidth);
      ptr += ScreenWidth;
   }
   ResumeScroller();
   return 0;
}

int  InvertScreenContext(void)
{
   unsigned char *ptr = ScreenContext;
   int  x, y;
   char c;

   if (!ptr) return 4;

   for(y=0; y<ScreenHeight; y++)
     for (x=0; x<ScreenWidth; x++)
     {  c = *ptr;
        if (c) *ptr = ~c;
        ptr++;
     }

   return 0;

}

void DeleteScreenContext(void)
{  if (ScreenContext)
   {   free(ScreenContext);
       ScreenContext = NULL;
   }
}

//-------------------------------------------------------------
int InvertPalette(void)
{  int Pal[45];
   int   i;

   if (VideoChipset == EGA) return 4;

   vga_getpalvec(1, 15, Pal);

   for (i=0; i<45; i++) Pal[i] ^= 0xff;

   vga_setpalvec(1, 15, Pal);
   return 0;
}

int SetStdPlette(void)
{   int Pal[48];
    int *ptr = Pal;
    int i;


   if (VideoChipset == EGA) return 4;

   // Dark colours
   for (i=0; i<7; i++)
   {   *ptr++ = (i & 4) ? 31 : 0;
       *ptr++ = (i & 2) ? 31 : 0;
       *ptr++ = (i & 1) ? 31 : 0;
   }

   // Colour 7 entry (light grey)
   // *ptr++ = *ptr++ = *ptr++ = 48;
       memset(ptr, 48, 3); ptr+= 3;

   // Colour 8 entry (dark grey)
   // *ptr++ = *ptr++ = *ptr++ = 31;
      memset(ptr, 31, 3); ptr+= 3;

   // Light colours
   for (i=9; i<16; i++)
   {   *ptr++ = (i & 4) ? 63 : 0;
       *ptr++ = (i & 2) ? 63 : 0;
       *ptr++ = (i & 1) ? 63 : 0;
   }

   vga_setpalvec(0, 16, Pal);
   return 0;
}

//-------------------------------------------------------------
void DrawSolidRect(RECT *rc, int colour)
{    char Buf1[100];
     int y;
     char *Buffer;

     int tox = rc->right;
     int fromx = rc->left;
     int wid = tox-fromx;

     int toy = rc->bottom;


     if (wid > 100) Buffer = (char *) malloc(wid);
              else  Buffer = Buf1;

     memset(Buffer, colour, wid);

     for(y= rc->top; y<toy; y++)
         setrow(fromx, y, wid, Buffer);

     if (wid > 100) free (Buffer);
}

void DrawRectWithFrame(RECT *rc, int flag )
{    int x, y, y1, off, i;
     int wid = rc->right - rc->left + 16;
     int hei = rc->bottom - rc->top + 10;
     char *buffer = (char *) malloc(wid);


     if (!buffer) return;

     x = rc->left - 8;
     off = rc->top - 6;
     y = off;
     y1 = hei-y+2*off;

     // Fill buffer with background colour
     memset(buffer, BkgrColour, wid);

     for(i=0; i<3; i++)
     { if (y >=0)
           setrow(x, y, wid, buffer);
       if (y1 < ScreenHeight)
          setrow(x, y1, wid, buffer);
       y++; y1--;
     }

     // Now draw two solid lines on both sides
     memset(buffer+4, FgrColour, wid-8);
     for (i=0; i<2; i++)
     { setrow(x, y, wid, buffer);
       setrow(x, y1, wid, buffer);
       y++; y1--;
     }


     // Now draw a frame if flag is set
     // Otherwise draw bottom only
     memset (buffer+7, BkgrColour, wid-14);
     i = flag ?  0 : hei-12;
     y = rc->top - 1 + i;

     for (; i<hei-9; i++)
     {    setrow(x, y, wid, buffer);
            y++;
     }


     free(buffer);
}

void DrawFramedRect(RECT *rc)
{   DrawRectWithFrame(rc, 1); }

void EncompassFramedRect(RECT *rc)
{   DrawRectWithFrame(rc, 0); }




//-------------------------------------------------------------
// Draw patterned rectangle
void DrawPatRect(RECT *rc, char pattern[8], int offset)
{
     char buffer[8];
     char patbyte;
     unsigned char mask;
     int fromx = rc->left, tox = rc->right;
     int fromy = rc->top, toy = rc->bottom;
     int x, y, i;

     if (toy > ScreenHeight) toy = ScreenHeight;

     for (i=0; i<8; i++)
     {  // Proceed pattern byte;

        patbyte = pattern[(i+offset)%8];
        mask = 0x80;

        for (y=0; y<8; y++)
        {  buffer[y] = (patbyte & mask) ?  FgrColour: BkgrColour;
           mask = (mask>>1) & 0x7f;
        }


        // Spread buffer wherever needed
        for (y=fromy+i; y<toy; y+=8)

           for (x=fromx; x<tox; x+=8)
            setrow(x, y, 8, buffer);

     }

}

void DrawBkgrPattern(RECT *rc)
{
   DrawPatRect(rc, FillPats[Level], 0);
}


//-------------------------------------------------------------
void DrawBoard(BOARD brd[][BRD_WIDTH],
          int fromx, int tox, int fromy, int toy)
{    int  y;

     if (fromy < 0) fromy = 0;
     if (toy > BRD_HEIGHT) toy = BRD_HEIGHT;

     if (fromx < 0) fromx = 0;
     if (tox > BRD_WIDTH) tox = BRD_WIDTH;

     for (y=fromy; y<toy; y++)
       DrawBoardRow(brd[y], fromx, tox,  y, 0);

     if (Demo) drawrobbrd(brd, fromx, tox, fromy, toy);

}


//-------------------------------------------------------------
static void DrawBackground(void)
{   RECT rc;
    FgrColour = BkgrPatColour[Level];

    rc.top = 0;
    rc.left = 0;
    rc.right = ScreenWidth;
    rc.bottom = ScreenHeight;
    DrawBkgrPattern(&rc);


    rc.top = BoardY;
    rc.left = BoardX;
    rc.right = BoardX + BoardW;
    rc.bottom = BoardY + BoardH;
    DrawFramedRect(&rc);

}

//-------------------------------------------------------------

void DrawAll(BOARD brd[][BRD_WIDTH])
{
    AllocScreenContext();
    DrawBackground();
    if (JoyEnabled || Demo) PrepareJoyWindow();
    PrepareScroller();
    DrawBoard(brd, 0, ScreenWidth, 0, ScreenHeight);
    PrepareStatWindow();
    if (ShowNext) PrepareNextWindow();
    if (ScreenHeight>350)
    {  showImage(&tetris_Hdr, HdrX, HdrY+8, 0);
       showSqueezedImage(&Intelligent_Hdr, HdrX-6, 2, 0, 2, 2);
    }
    drawpic();
    ScreenContextToScreen();
    DeleteScreenContext();
}

//-----------------------------------------------------------
/*
void DrawPatLine(int x, int y, int width, int pat, int fgr, int bkgr)
{  char pattern[8];
   int  i, m;

   m = 0x80;
   for(i=0; i<8; i++)
   {    pattern[i] = (pat & m) ? fgr : bkgr;
        m >>= 1;
   }

   i = 0;
   while (i<width)
   {  m = 8;
      if (i+m > width) m = width-i;
      setrow(x+i, y, m, pattern);
      i += m;
   }

}

*/
void DrawSolidLine(int x1, int y1, int x2, int y2, int width, int Colour)
{

     vga_setcolor(Colour);
     
     vga_drawline(x1, y1, x2, y2);
}

